import re
from importlib import import_module
from typing import Tuple

PRETTY_REGEX = re.compile(r'([\w ]*?) *<(.*)> *')

# max length for domain name labels is 63 characters per RFC 1034
EMAIL_REGEX = re.compile(
    r'[^\s@\u0000-\u0020"\'`,]+@'
    r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z0-9]{2,63})',
    re.I
)


def validate_email(value) -> Tuple[str, str]:
    """
    Brutally simple email address validation. Note unlike most email address validation
    * raw ip address (literal) domain parts are not allowed.
    * "John Doe <local_part@domain.com>" style "pretty" email addresses are processed
    * the local part check is extremely basic. This raises the possibility of unicode spoofing, but no better
        solution is really possible.
    * spaces are striped from the beginning and end of addresses but no error is raised

    See RFC 5322 but treat it with suspicion, there seems to exist no universally acknowledged test for a valid email!
    """
    m = PRETTY_REGEX.fullmatch(value)
    if m:
        name, value = m.groups()
    else:
        name = None

    email = value.strip()
    if not EMAIL_REGEX.fullmatch(email):
        raise ValueError('Email address is not valid')
    return name or email[:email.index('@')], email.lower()


def _rfc_1738_quote(text):
    return re.sub(r'[:@/]', lambda m: '%{:X}'.format(ord(m.group(0))), text)


def make_dsn(*,
             driver: str,
             user: str=None,
             password: str=None,
             host: str=None,
             port: str=None,
             name: str=None,
             query: str=None):
    """
    Create a DSN from from connection settings.

    Stolen approximately from sqlalchemy/engine/url.py:URL.
    """
    s = driver + '://'
    if user is not None:
        s += _rfc_1738_quote(user)
        if password is not None:
            s += ':' + _rfc_1738_quote(password)
        s += '@'
    if host is not None:
        if ':' in host:
            s += '[{}]'.format(host)
        else:
            s += host
    if port is not None:
        s += ':{}'.format(int(port))
    if name is not None:
        s += '/' + name
    query = query or {}
    if query:
        keys = list(query)
        keys.sort()
        s += '?' + '&'.join('{}={}'.format(k, query[k]) for k in keys)
    return s


def import_string(dotted_path):
    """
    Stolen approximately from django. Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.strip(' ').rsplit('.', 1)
    except ValueError as e:
        raise ImportError(f'"{dotted_path}" doesn\'t look like a module path') from e

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute') from e



from collections import OrderedDict
from types import FunctionType

from .exceptions import Error, ValidationError
from .fields import Field
from .validators import dict_validator


class BaseConfig:
    min_anystr_length = 0
    max_anystr_length = 2 ** 16
    min_number_size = -2 ** 64
    max_number_size = 2 ** 64
    raise_exception = True
    validate_all = False
    ignore_extra = True


def inherit_config(self_config, parent_config) -> BaseConfig:
    if not self_config:
        return parent_config
    for k, v in parent_config.__dict__.items():
        if not (k.startswith('_') or hasattr(self_config, k)):
            setattr(self_config, k, v)
    return self_config


TYPE_BLACKLIST = FunctionType, property, type, classmethod, staticmethod


class MetaModel(type):
    @classmethod
    def __prepare__(mcs, *args, **kwargs):
        return OrderedDict()

    def __new__(mcs, name, bases, namespace):
        fields = OrderedDict()
        config = BaseConfig
        for base in reversed(bases):
            if issubclass(base, BaseModel) and base != BaseModel:
                fields.update(base.__fields__)
                config = inherit_config(base.config, config)

        annotations = namespace.get('__annotations__')
        config = inherit_config(namespace.get('Config'), config)
        class_validators = {n: f for n, f in namespace.items()
                            if n.startswith('validate_') and isinstance(f, FunctionType)}

        for var_name, value in namespace.items():
            if var_name.startswith('_') or isinstance(value, TYPE_BLACKLIST):
                continue
            field = Field.infer(
                name=var_name,
                value=value,
                annotation=annotations and annotations.get(var_name),
                class_validators=class_validators,
            )
            fields[var_name] = field
        namespace.update(
            config=config,
            __fields__=fields,
        )
        return super().__new__(mcs, name, bases, namespace)


class Missing(ValueError):
    pass


class Extra(ValueError):
    pass


MISSING = Missing('field required')
MISSING_ERROR = Error(MISSING, None, None, None)
EXTRA_ERROR = Error(Extra('extra fields not permitted'), None, None, None)


class BaseModel(metaclass=MetaModel):
    # populated by the metaclass, defined here to help IDEs only
    __fields__ = {}

    def __init__(self, **values):
        self.__values__ = {}
        self.__errors__ = OrderedDict()
        self._process_values(values)

    @property
    def values(self):
        return self.__values__

    @property
    def fields(self):
        return self.__fields__

    @property
    def errors(self):
        return self.__errors__

    @classmethod
    def get_validators(cls):
        yield dict_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return cls(**value)

    def _process_values(self, values):
        for name, field in self.__fields__.items():
            value = values.get(name, MISSING)
            self._process_value(name, field, value)

        if not self.config.ignore_extra:
            extra = values.keys() - self.__fields__.keys()
            if extra:
                for field in sorted(extra):
                    self.__errors__[field] = EXTRA_ERROR

        if self.config.raise_exception and self.__errors__:
            raise ValidationError(self.__errors__)

    def _process_value(self, name, field, value):
        if value is MISSING:
            if self.config.validate_all or field.validate_always:
                value = field.default
            else:
                if field.required:
                    self.__errors__[name] = MISSING_ERROR
                else:
                    self.__values__[name] = field.default
                    # could skip this if the attributes equals field.default, would it be quicker?
                    setattr(self, name, field.default)
                return

        value, errors = field.validate(value, self)
        if errors:
            self.__errors__[name] = errors
        self.__values__[name] = value
        setattr(self, name, value)

    def __iter__(self):
        # so `dict(model)` works
        yield from self.__values__.items()

    def __repr__(self):
        return f'<{self}>'

    def __str__(self):
        return '{} {}'.format(self.__class__.__name__, ' '.join('{}={!r}'.format(k, v)
                                                                for k, v in self.__values__.items()))



import json
from collections import OrderedDict, namedtuple


def type_json(type_: type):
    if type_:
        try:
            return type_.__name__
        except AttributeError:
            # happens with unions
            return str(type_)


Error = namedtuple('Error', ['exc', 'validator', 'track_type', 'index'])


def jsonify_errors(e):
    if not e:
        return e
    elif isinstance(e, Error):
        return {
            'error_type': e.exc.__class__.__name__,
            'error_msg': str(e.exc),
            'validator': e.validator and e.validator.__qualname__,
            'track': type_json(e.track_type),
            'index': e.index,
        }
    elif isinstance(e, OrderedDict):
        return OrderedDict([(k, jsonify_errors(v)) for k, v in e.items()])
    else:
        return [jsonify_errors(e_) for e_ in e]


class ValidationError(ValueError):
    def __init__(self, errors):
        self.errors = errors
        e_count = len(self.errors)
        s = '' if e_count == 1 else 's'
        self.message = f'{e_count} error{s} validating input'
        self.errors_jsonable = jsonify_errors(errors)
        super().__init__(f'{self.message}: {self.json()}')

    def json(self, indent=None):
        return json.dumps(self.errors_jsonable, indent=indent, sort_keys=True)


class ConfigError(RuntimeError):
    pass



import inspect
from collections import OrderedDict
from enum import IntEnum
from typing import Any, List, Mapping, Sequence, Type, Union  # noqa

from .exceptions import ConfigError, Error, type_json
from .validators import NoneType, find_validator, not_none_validator


class ValidatorSignature(IntEnum):
    JUST_VALUE = 1
    VALUE_KWARGS = 2
    BOUND_METHOD = 3


class Shape(IntEnum):
    SINGLETON = 1
    SEQUENCE = 2
    MAPPING = 3


class Field:
    __slots__ = ('type_', 'key_type_', 'validator_tracks', 'key_validator_tracks', 'track_count', 'default',
                 'required', 'name', 'description', 'info', 'validate_always', 'allow_none', 'shape')

    def __init__(
            self, *,
            name: str=None,
            type_: Type,
            class_validators: dict=None,
            default: Any=None,
            required: bool=False,
            description: str=None):

        self.name: str = name
        self.type_: type = type_
        self.key_type_: type = None
        self.validate_always: bool = getattr(self.type_, 'validate_always', False)
        self.validator_tracks: List[ValidatorRoute] = []
        self.key_validator_tracks: List[ValidatorRoute] = []
        self.default: Any = default
        self.required: bool = required
        self.description: str = description
        self.allow_none: bool = False
        self.shape: Shape = Shape.SINGLETON
        self._prepare(class_validators)

    @classmethod
    def infer(cls, *, name, value, annotation, class_validators):
        required = value == Ellipsis
        return cls(
            type_=annotation,
            class_validators=class_validators,
            default=None if required else value,
            required=required,
            name=name
        )

    def _prepare(self, class_validators):
        if self.default and self.type_ is None:
            self.type_ = type(self.default)

        if self.type_ is None:
            raise ConfigError(f'unable to infer type for attribute "{self.name}"')

        # typing interface is horrible, we have to do some ugly checks
        origin = getattr(self.type_, '__origin__', None)
        if origin not in (None, Union):
            if issubclass(origin, Sequence):
                self.type_ = self.type_.__args__[0]
                self.shape = Shape.SEQUENCE
            else:
                assert issubclass(origin, Mapping)
                self.key_type_ = self.type_.__args__[0]
                self.type_ = self.type_.__args__[1]
                self.shape = Shape.MAPPING

        self.validator_tracks = self._populate_validator_tracks(self.type_, class_validators)
        if self.key_type_:
            self.key_validator_tracks = self._populate_validator_tracks(self.key_type_, class_validators, 'key_')
        validators = {
            type_json(r.type_): [v[1].__qualname__ for v in r.validators] for r in self.validator_tracks
        }
        if len(validators) == 1:
            validators = list(validators.values())[0]
        self.info = OrderedDict([
            ('type', type_json(self.type_)),
            ('default', self.default),
            ('required', self.required),
            ('validators', validators)
        ])
        if self.required:
            self.info.pop('default')
        if self.description:
            self.info['description'] = self.description

    def _populate_validator_tracks(self, type_, class_validators, prefix=''):
        class_validators = class_validators or {}
        override_validator = class_validators.get(f'validate_{prefix}{self.name}_override')
        if override_validator:
            tracks = [ValidatorRoute(self.type_, override_validator)]
        else:
            tracks = []
            if getattr(self.type_, '__origin__', None) is Union:
                types = type_.__args__
            else:
                types = [type_]

            for type_ in types:
                if type_ is NoneType:
                    self.allow_none = True
                else:
                    tracks.append(ValidatorRoute(type_))

        for track in tracks:
            track.prepend(class_validators.get(f'validate_{prefix}{self.name}_pre'))
            track.append(class_validators.get(f'validate_{prefix}{self.name}'))
            track.append(class_validators.get(f'validate_{prefix}{self.name}_post'))
            track.freeze(none_track=self.allow_none)
        return tracks

    def validate(self, v, model):
        if self.allow_none and v is None:
            return None, None

        if self.shape == Shape.SINGLETON:
            return self._validate_singleton(self.validator_tracks, v, model)
        elif self.shape == Shape.SEQUENCE:
            return self._validate_sequence(v, model)
        else:
            return self._validate_mapping(v, model)

    def _validate_sequence(self, v, model):
        result, errors = [], []
        try:
            v_iter = enumerate(v)
        except TypeError as exc:
            return v, Error(exc, iter, None, None)
        for i, v_ in v_iter:
            single_result, single_errors = self._validate_singleton(self.validator_tracks, v_, model, i)
            if errors or single_errors:
                errors.append(single_errors)
            else:
                result.append(single_result)
        if errors:
            return v, errors
        else:
            return result, None

    def _validate_mapping(self, v, model):
        if isinstance(v, dict):
            v_iter = v
        else:
            try:
                v_iter = dict(v)
            except TypeError as exc:
                return v, Error(exc, dict, None, None)

        result, errors = {}, []
        for k, v_ in v_iter.items():
            key_result, key_errors = self._validate_singleton(self.key_validator_tracks, k, model)
            value_result, value_errors = self._validate_singleton(self.validator_tracks, v_, model, k)

            if errors or value_errors or key_errors:
                errors.append([key_errors, value_errors])
            else:
                result[key_result] = value_result
        if errors:
            return v, errors
        else:
            return result, None

    def _validate_singleton(self, tracks, v, model, index=None):
        result, errors = ..., []
        for track in tracks:
            value, exc, validator = track.validate(v, model, self)
            if exc:
                errors.append(Error(exc, validator, track.type_, index))
            elif isinstance(v, track.type_):
                # exact match: return immediately
                return value, None
            else:
                result = value
        if result is not ...:
            return result, None
        else:
            return v, errors[0] if len(tracks) == 1 else errors

    def __repr__(self):
        return f'<Field {self}>'

    def __str__(self):
        return f'{self.name}: ' + ', '.join(f'{k}={v!r}' for k, v in self.info.items())


class ValidatorRoute:
    __slots__ = 'type_', 'validators', '_none_track'

    def __init__(self, type_, *validators):
        self.type_ = type_
        self.validators = list(validators or self._find_validator())
        self._none_track = None

    def prepend(self, validator):
        self.validators.insert(0, validator)

    def append(self, validator):
        self.validators.append(validator)

    def freeze(self, none_track):
        self._none_track = none_track
        tmp_validators = []
        for validator in self.validators:
            if not validator:
                continue
            if none_track and validator is not_none_validator:
                continue
            signature = self._get_validator_signature(validator)
            tmp_validators.append((signature, validator))

        self.validators = tuple(tmp_validators)

    def validate(self, v, model, field):
        for signature, validator in self.validators:
            try:
                if signature == ValidatorSignature.JUST_VALUE:
                    v = validator(v)
                elif signature == ValidatorSignature.VALUE_KWARGS:
                    v = validator(v, model=model, field=field)
                else:
                    v = validator(model, v)
            except (ValueError, TypeError, ImportError) as e:
                return v, e, validator
        return v, None, None

    def _find_validator(self):
        get_validators = getattr(self.type_, 'get_validators', None)
        if get_validators:
            return list(get_validators())
        return find_validator(self.type_)

    @staticmethod
    def _get_validator_signature(validator):
        try:
            signature = inspect.signature(validator)
        except ValueError:
            # happens on builtins like float
            return ValidatorSignature.JUST_VALUE

        # bind here will raise a TypeError so:
        # 1. we can deal with it before validation begins
        # 2. (more importantly) it doesn't get confused with a TypeError when evaluating the validator
        try:
            if list(signature.parameters)[0] == 'self':
                signature.bind(object(), 1)
                return ValidatorSignature.BOUND_METHOD
            elif len(signature.parameters) == 1:
                signature.bind(1)
                return ValidatorSignature.JUST_VALUE
            else:
                signature.bind(1, model=2, field=3)
                return ValidatorSignature.VALUE_KWARGS
        except TypeError as e:
            raise ConfigError(f'Invalid signature for validator {validator}: {signature}, should be: '
                              f'(value) or (value, *, model, field)') from e



from collections import OrderedDict
from datetime import date, datetime, time, timedelta
from enum import Enum
from pathlib import Path

from .datetime_parse import parse_date, parse_datetime, parse_duration, parse_time
from .exceptions import ConfigError

NoneType = type(None)


def not_none_validator(v):
    if v is None:
        raise TypeError('None is not an allow value')
    return v


def str_validator(v) -> str:
    if isinstance(v, (str, NoneType)):
        return v
    elif isinstance(v, bytes):
        return v.decode()
    return str(v)


def bytes_validator(v) -> bytes:
    if isinstance(v, (bytes, NoneType)):
        return v
    return str_validator(v).encode()


BOOL_STRINGS = {
    '1',
    'TRUE',
    'ON',
    'YES',
}


def bool_validator(v) -> bool:
    if isinstance(v, bool):
        return v
    if isinstance(v, bytes):
        v = v.decode()
    if isinstance(v, str):
        return v.upper() in BOOL_STRINGS
    return bool(v)


def number_size_validator(v, model, **kwargs):
    if model.config.min_number_size <= v <= model.config.max_number_size:
        return v
    raise ValueError(f'size not in range {model.config.min_number_size} to {model.config.max_number_size}')


def anystr_length_validator(v, model, **kwargs):
    if v is None or model.config.min_anystr_length <= len(v) <= model.config.max_anystr_length:
        return v
    raise ValueError(f'length not in range {model.config.max_anystr_length} to {model.config.max_anystr_length}')


def ordered_dict_validator(v) -> OrderedDict:
    if isinstance(v, OrderedDict):
        return v
    return OrderedDict(v)


def dict_validator(v) -> dict:
    if isinstance(v, dict):
        return v
    return dict(v)


def list_validator(v) -> list:
    if isinstance(v, list):
        return v
    return list(v)


def enum_validator(v, field, **kwargs) -> Enum:
    return field.type_(v)


# order is important here, for example: bool is a subclass of int, datetime is a subclass of date
_VALIDATORS = [
    (str, [not_none_validator, str_validator, anystr_length_validator]),
    (bytes, [not_none_validator, bytes_validator, anystr_length_validator]),

    (Enum, [enum_validator]),

    (bool, [bool_validator]),
    (int, [int, number_size_validator]),
    (float, [float, number_size_validator]),

    (Path, [Path]),

    (datetime, [parse_datetime]),
    (date, [parse_date]),
    (time, [parse_time]),
    (timedelta, [parse_duration]),

    (OrderedDict, [ordered_dict_validator]),
    (dict, [dict_validator]),
    (list, [list_validator]),
]


def find_validator(type_):
    for val_type, validators in _VALIDATORS:
        if issubclass(type_, val_type):
            return validators
    raise ConfigError(f'no validator found for {type_}')



import re
from typing import Optional, Type, Union

from .utils import import_string, make_dsn, validate_email
from .validators import str_validator

__all__ = [
    'NoneStr',
    'NoneBytes',
    'StrBytes',
    'NoneStrBytes',
    'ConstrainedStr',
    'constr',
    'EmailStr',
    'NameEmail',
    'Module',
    'DSN',
]

NoneStr = Optional[str]
NoneBytes = Optional[bytes]
StrBytes = Union[str, bytes]
NoneStrBytes = Optional[StrBytes]


class ConstrainedStr(str):
    min_length: int = None
    max_length: int = None
    curtail_length: int = None
    regex = None

    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        l = len(value)
        if cls.min_length and l < cls.min_length:
            raise ValueError(f'length less than minimum allowed: {cls.min_length}')

        if cls.curtail_length:
            if l > cls.curtail_length:
                value = value[:cls.curtail_length]
        elif cls.max_length and l > cls.max_length:
            raise ValueError(f'length greater than maximum allowed: {cls.max_length}')

        if cls.regex:
            if not cls.regex.match(value):
                raise ValueError(f'string does not match regex "{cls.regex.pattern}"')
        return value


class EmailStr(str):
    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return validate_email(value)[1]


class NameEmail:
    __slots__ = 'name', 'email'

    def __init__(self, name, email):
        self.name = name
        self.email = email

    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return cls(*validate_email(value))


def constr(*, min_length=0, max_length=2**16, curtail_length=None, regex=None) -> Type[str]:
    # use kwargs then define conf in a dict to aid with IDE type hinting
    namespace = dict(
        min_length=min_length,
        max_length=max_length,
        curtail_length=curtail_length,
        regex=regex and re.compile(regex)
    )
    return type('ConstrainedStrValue', (ConstrainedStr,), namespace)


class Module:
    validate_always = True

    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return import_string(value)


class DSN(str):
    prefix = 'db_'
    validate_always = True

    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value, model, **kwarg):
        if value:
            return value
        d = model.__values__
        kwargs = {f: d.get(cls.prefix + f) for f in ('driver', 'user', 'password', 'host', 'port', 'name', 'query')}
        if kwargs['driver'] is None:
            raise ValueError(f'"{cls.prefix}driver" field may not be missing or None')
        return make_dsn(**kwargs)


# TODO, JsonEither, JsonList, JsonDict
